#include "GeoMesh.h"

using namespace Geometry;
using namespace std;

//---------------------------------------------------------------------------
//	Cosntructor.
//---------------------------------------------------------------------------
Mesh::Mesh():
	mVertexBuffer(0), mSubMeshCount(0), mSubMesh(0), hasSkeleton(false)
{
	mMeshBounds.maxX					=	0.0;
	mMeshBounds.maxY					=	0.0;
	mMeshBounds.maxZ					=	0.0;
	mMeshBounds.minX					=	0.0;
	mMeshBounds.minY					=	0.0;
	mMeshBounds.minZ					=	0.0;
	mMeshBounds.scaleFactor	=	0.0;
}

//---------------------------------------------------------------------------
//	Destroyer.
//---------------------------------------------------------------------------
Mesh::~Mesh()
{
	delete[] mSubMesh;
	delete mVertexBuffer;
}


//---------------------------------------------------------------------------
//	Copy constructor.
//---------------------------------------------------------------------------
Mesh::Mesh(const Mesh &objmesh) 
{
	// constructor copia
	mVertexBuffer	=	new VertexBuffer();
	mSubMeshCount	=	objmesh.mSubMeshCount;
	mSubMesh			=	new SubMesh[objmesh.mSubMeshCount];

	//	Fill up bounding box settings.
	mMeshBounds.maxX					=	objmesh.mMeshBounds.maxX;
	mMeshBounds.maxY					=	objmesh.mMeshBounds.maxY;
	mMeshBounds.maxZ					=	objmesh.mMeshBounds.maxZ;
	mMeshBounds.minX					=	objmesh.mMeshBounds.minX;
	mMeshBounds.minY					=	objmesh.mMeshBounds.minY;
	mMeshBounds.minZ					=	objmesh.mMeshBounds.minZ;
	mMeshBounds.radius				=	objmesh.mMeshBounds.radius;
	mMeshBounds.scaleFactor		=	objmesh.mMeshBounds.scaleFactor;
	
	//	For each submesh.
	for(size_t i = 0; i < objmesh.mSubMeshCount; i++)
	{
		mSubMesh[i].mSharedVertexBuffer					=	false;
		mSubMesh[i].mVertexBuffer								=	new VertexBuffer();
		mSubMesh[i].mVertexBuffer->mPosition		=	new Vector3[objmesh.mSubMesh[i].mVertexBuffer->mVertexCount];
		mSubMesh[i].mVertexBuffer->mNormal			=	new Vector3[objmesh.mSubMesh[i].mVertexBuffer->mVertexCount];
		mSubMesh[i].mVertexBuffer->mTexCoords 	= new Vector2[objmesh.mSubMesh[i].mVertexBuffer->mVertexCount];
		mSubMesh[i].mVertexBuffer->mVertexCount =	objmesh.mSubMesh[i].mVertexBuffer->mVertexCount;
		mSubMesh[i].mVertexBuffer->mVertexInfo	=	objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo;
		mSubMesh[i].mType												=	objmesh.mSubMesh[i].mType;
		strcpy(mSubMesh[i].mMaterialName,objmesh.mSubMesh[i].mMaterialName);
		
		if(objmesh.mSubMesh[i].mSharedVertexBuffer) 
		{
			for(size_t s	=	0;
					s<objmesh.mSubMesh[i].mVertexBuffer->mVertexCount;
					s++)
			{
				//	Copy mPosition.
				if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & VERTEX_POSITION)
				{
					mSubMesh[i].mVertexBuffer->mPosition[s].x = objmesh.mVertexBuffer->mPosition[s].x;
					mSubMesh[i].mVertexBuffer->mPosition[s].y = objmesh.mVertexBuffer->mPosition[s].y;
					mSubMesh[i].mVertexBuffer->mPosition[s].z = objmesh.mVertexBuffer->mPosition[s].z;
				}
				
				//	Copy mNormal.
				if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & VERTEX_NORMAL)
				{
					mSubMesh[i].mVertexBuffer->mNormal[s].x = objmesh.mVertexBuffer->mNormal[s].x;
					mSubMesh[i].mVertexBuffer->mNormal[s].y = objmesh.mVertexBuffer->mNormal[s].y;
					mSubMesh[i].mVertexBuffer->mNormal[s].z = objmesh.mVertexBuffer->mNormal[s].z;
				}

				//	Copy mTexCoords.
				if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & VERTEX_TEXCOORDS)
				{
					mSubMesh[i].mVertexBuffer->mTexCoords[s].x = objmesh.mVertexBuffer->mTexCoords[s].x;
					mSubMesh[i].mVertexBuffer->mTexCoords[s].y = objmesh.mVertexBuffer->mTexCoords[s].y;
				}
			}
			objmesh.mSubMesh[i].mSharedVertexBuffer = false;
		}
		else
		{
			for(size_t s = 0; s < objmesh.mSubMesh[i].mVertexBuffer->mVertexCount; s++)
			{
				//	Copy mPosition.
				if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & VERTEX_POSITION)
				{
					mSubMesh[i].mVertexBuffer->mPosition[s].x = objmesh.mSubMesh[i].mVertexBuffer->mPosition[s].x;
					mSubMesh[i].mVertexBuffer->mPosition[s].y = objmesh.mSubMesh[i].mVertexBuffer->mPosition[s].y;
					mSubMesh[i].mVertexBuffer->mPosition[s].z = objmesh.mSubMesh[i].mVertexBuffer->mPosition[s].z;
				}
				
				//Copiamos mNormal
				if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & VERTEX_NORMAL)
				{
					mSubMesh[i].mVertexBuffer->mNormal[s].x = objmesh.mSubMesh[i].mVertexBuffer->mNormal[s].x;
					mSubMesh[i].mVertexBuffer->mNormal[s].y = objmesh.mSubMesh[i].mVertexBuffer->mNormal[s].y;
					mSubMesh[i].mVertexBuffer->mNormal[s].z = objmesh.mSubMesh[i].mVertexBuffer->mNormal[s].z;
				}

				//Copiamos mTexCoords
				if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & VERTEX_TEXCOORDS)
				{
					mSubMesh[i].mVertexBuffer->mTexCoords[s].x = objmesh.mSubMesh[i].mVertexBuffer->mTexCoords[s].x;
					mSubMesh[i].mVertexBuffer->mTexCoords[s].y = objmesh.mSubMesh[i].mVertexBuffer->mTexCoords[s].y;
				}
			}

		}
		
		// Copy indices.
		mSubMesh[i].mIndexCount	=	objmesh.mSubMesh[i].mIndexCount;
		mSubMesh[i].mIndex			=	new Index[mSubMesh[i].mIndexCount];
		
		memcpy(mSubMesh[i].mIndex,objmesh.mSubMesh[i].mIndex,objmesh.mSubMesh[i].mIndexCount*sizeof(Index));

		// Copy strips.
		__w64 int offset	=	0; // offset between memory positions.
		mSubMesh[i].mStripCount=objmesh.mSubMesh[i].mStripCount;
		if (objmesh.mSubMesh[i].mStripCount>0) 
		{
			mSubMesh[i].mStrip = new Index*[objmesh.mSubMesh[i].mStripCount];

			offset	= &(mSubMesh[i].mIndex[0]) - &(objmesh.mSubMesh[i].mIndex[0]);
			for (size_t j=0;j<objmesh.mSubMesh[i].mStripCount;j++)
			{
				mSubMesh[i].mStrip[j]=objmesh.mSubMesh[i].mStrip[j] + offset;
			}
		}

		//	Copy submesh bones.
		if (!objmesh.mSubMesh[i].mBones.empty())
		{
			for (unsigned int j = 0; j < objmesh.mSubMesh[i].mBones.size(); j++)
			{
				mSubMesh[i].mBones.push_back(objmesh.mSubMesh[i].mBones[j]);
			}
		}
	}
	
	//	Copy skeleton name.
	if (objmesh.hasSkeleton)
	{
		hasSkeleton	=	true;
		
		strcpy(mSkeletonName,objmesh.mSkeletonName);
	}

	//	Copy mesh bones.
	if (!objmesh.mBones.empty())
	{
		for (unsigned int j = 0; j < objmesh.mBones.size(); j++)
		{
			mBones.push_back(objmesh.mBones[j]);
		}
	}

}

//---------------------------------------------------------------------------
//	Assignment operator.
//---------------------------------------------------------------------------
Mesh &Mesh::operator =(const Geometry::Mesh &objmesh)
{
	// Operador de asignacin
	bool copiados	=	false; // indica si los vrtices compartidos han sido copiados
	mVertexBuffer	=	new VertexBuffer();
	mSubMeshCount	=	objmesh.mSubMeshCount;
	mSubMesh			=	new Geometry::SubMesh[objmesh.mSubMeshCount];
	
	//	Fill up bounding box settings.
	mMeshBounds.maxX					=	objmesh.mMeshBounds.maxX;
	mMeshBounds.maxY					=	objmesh.mMeshBounds.maxY;
	mMeshBounds.maxZ					=	objmesh.mMeshBounds.maxZ;
	mMeshBounds.minX					=	objmesh.mMeshBounds.minX;
	mMeshBounds.minY					=	objmesh.mMeshBounds.minY;
	mMeshBounds.minZ					=	objmesh.mMeshBounds.minZ;
	mMeshBounds.radius				=	objmesh.mMeshBounds.radius;
	mMeshBounds.scaleFactor		=	objmesh.mMeshBounds.scaleFactor;

	//	For each submesh.
	for(size_t i = 0; i < objmesh.mSubMeshCount; i++)
	{
		mSubMesh[i].mSharedVertexBuffer					=	objmesh.mSubMesh[i].mSharedVertexBuffer; //.false;
		mSubMesh[i].mVertexBuffer								=	new Geometry::VertexBuffer();
		mSubMesh[i].mVertexBuffer->mPosition		=	new Geometry::Vector3[objmesh.mSubMesh[i].mVertexBuffer->mVertexCount];
		mSubMesh[i].mVertexBuffer->mNormal			=	new Geometry::Vector3[objmesh.mSubMesh[i].mVertexBuffer->mVertexCount];
		mSubMesh[i].mVertexBuffer->mTexCoords		=	new Geometry::Vector2[objmesh.mSubMesh[i].mVertexBuffer->mVertexCount];
		mSubMesh[i].mVertexBuffer->mVertexCount	=	objmesh.mSubMesh[i].mVertexBuffer->mVertexCount;
		mSubMesh[i].mVertexBuffer->mVertexInfo	=	objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo;
		mSubMesh[i].mType					=	objmesh.mSubMesh[i].mType;
		strcpy(mSubMesh[i].mMaterialName,objmesh.mSubMesh[i].mMaterialName);

		if (objmesh.mSubMesh[i].mSharedVertexBuffer && !copiados)
		{
			mVertexBuffer	=	mSubMesh[i].mVertexBuffer;
			copiados			=	true;
		}

		for(size_t s = 0; s < objmesh.mSubMesh[i].mVertexBuffer->mVertexCount; s++)
		{
			//Copiamos mPosition
			//if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & Geometry::VERTEX_POSITION)
			//{
				mSubMesh[i].mVertexBuffer->mPosition[s].x = objmesh.mSubMesh[i].mVertexBuffer->mPosition[s].x;
				mSubMesh[i].mVertexBuffer->mPosition[s].y = objmesh.mSubMesh[i].mVertexBuffer->mPosition[s].y;
				mSubMesh[i].mVertexBuffer->mPosition[s].z = objmesh.mSubMesh[i].mVertexBuffer->mPosition[s].z;
			//}
			//Copiamos mNormal
			if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & Geometry::VERTEX_NORMAL)
			{
				mSubMesh[i].mVertexBuffer->mNormal[s].x = objmesh.mSubMesh[i].mVertexBuffer->mNormal[s].x;
				mSubMesh[i].mVertexBuffer->mNormal[s].y = objmesh.mSubMesh[i].mVertexBuffer->mNormal[s].y;
				mSubMesh[i].mVertexBuffer->mNormal[s].z = objmesh.mSubMesh[i].mVertexBuffer->mNormal[s].z;
			}

			//Copiamos mTexCoords
			if (objmesh.mSubMesh[i].mVertexBuffer->mVertexInfo & Geometry::VERTEX_TEXCOORDS)
			{
				mSubMesh[i].mVertexBuffer->mTexCoords[s].x = objmesh.mSubMesh[i].mVertexBuffer->mTexCoords[s].x;
				mSubMesh[i].mVertexBuffer->mTexCoords[s].y = objmesh.mSubMesh[i].mVertexBuffer->mTexCoords[s].y;
			}
		}

		// Copiar los indices
		mSubMesh[i].mIndexCount	=	objmesh.mSubMesh[i].mIndexCount;
		mSubMesh[i].mIndex			=	new Index[mSubMesh[i].mIndexCount];

		memcpy(	mSubMesh[i].mIndex,
						objmesh.mSubMesh[i].mIndex,
						objmesh.mSubMesh[i].mIndexCount*sizeof(Geometry::Index));

		// Copiar las tiras
		__w64 offset	=	0; // desplazamiento entre posiciones de memoria.
		mSubMesh[i].mStripCount	=	objmesh.mSubMesh[i].mStripCount;

		if (objmesh.mSubMesh[i].mStripCount>0) 
		{
			mSubMesh[i].mStrip	=	new Index*[objmesh.mSubMesh[i].mStripCount];
			offset	=	&(mSubMesh[i].mIndex[0]) - &(objmesh.mSubMesh[i].mIndex[0]);
			
			for (size_t j = 0; j < objmesh.mSubMesh[i].mStripCount; j++)
			{
				mSubMesh[i].mStrip[j]	=	objmesh.mSubMesh[i].mStrip[j] + offset;
			}
		}
		
		//	Copy submesh bones.
		if (!objmesh.mSubMesh[i].mBones.empty())
		{
			for (unsigned int j = 0; j < objmesh.mSubMesh[i].mBones.size(); j++)
			{
				mSubMesh[i].mBones.push_back(objmesh.mSubMesh[i].mBones[j]);
			}
		}
	}

	//	Copy skeleton name.
	if (objmesh.hasSkeleton)
	{
		hasSkeleton	=	true;
		
		strcpy(mSkeletonName,objmesh.mSkeletonName);
	}

	//	Copy mesh bones.
	if (!objmesh.mBones.empty())
	{
		for (unsigned int j = 0; j < objmesh.mBones.size(); j++)
		{
			mBones.push_back(objmesh.mBones[j]);
		}
	}

	return	*this;
}

//---------------------------------------------------------------------------
//	Load mesh.
//---------------------------------------------------------------------------
void Mesh::Load(Serializer &s)
{
	bool sharedVertexBuffer	= false;

	//Clear Data
	delete[] mSubMesh;  mSubMesh = 0;
	delete   mVertexBuffer; mVertexBuffer = 0;
	mSubMeshCount = 0;
	
	s.ReadArray(&sharedVertexBuffer,1);
	if (sharedVertexBuffer)
	{
		mVertexBuffer = new VertexBuffer;
		mVertexBuffer->Load(s);
	}
	s.ReadArray(&mSubMeshCount,1);
	mSubMesh = new SubMesh[mSubMeshCount];
	for(size_t i = 0; i < mSubMeshCount; ++i)
	{
		mSubMesh[i].Load(s);
		if (mSubMesh[i].mSharedVertexBuffer && sharedVertexBuffer)
		{
			mSubMesh[i].mVertexBuffer = mVertexBuffer;
		}
	}
}

//---------------------------------------------------------------------------
//	Save mesh.
//---------------------------------------------------------------------------
void Mesh::Save(Serializer &s)
{
	bool sharedVertexBuffer = (mVertexBuffer != 0);
	s.WriteArray(&sharedVertexBuffer, 1);
	if (sharedVertexBuffer)
	{
		mVertexBuffer->Save(s);
	}
	s.WriteArray(&mSubMeshCount,1);
	for(size_t i = 0; i < mSubMeshCount; ++i)
	{
		mSubMesh[i].Save(s);
	}
}

//---------------------------------------------------------------------------
//	Export to obj mesh.
//---------------------------------------------------------------------------
void Mesh::exportToOBJ(char *nomfich)
{
	// Genera un fichero obj con vrtices, tringulos

	std::ofstream obj(nomfich);
	obj << "begin" << std::endl;
	// Vrtices
	for (size_t i=0; i < mSubMeshCount; i++) 
	{
		for (size_t j=0; j < mSubMesh[i].mVertexBuffer->mVertexCount; j++) 
		{	
			obj << "v " <<  mSubMesh[i].mVertexBuffer->mPosition[j].x << " " <<
				mSubMesh[i].mVertexBuffer->mPosition[j].y << " " <<
				mSubMesh[i].mVertexBuffer->mPosition[j].z << " " << std::endl;
		}
	}


	// Caras
	for (size_t i=0; i < mSubMeshCount; i++) 
	{
		for (size_t j=0; j < mSubMesh[i].mIndexCount; j=j+3) 
		{	
			obj << "f " <<  mSubMesh[i].mIndex[j]+1 << " " << 
				mSubMesh[i].mIndex[j+1]+1 << " " << 
				mSubMesh[i].mIndex[j+2]+1 << std::endl;
		}
	}

	obj << "end" << std::endl;
	obj.close();
}

//---------------------------------------------------------------------------
// Transform to shared vertex mesh.
//---------------------------------------------------------------------------
Mesh *Mesh::toSharedVertex() 
{
	// Move all vertex to the shared vertex buffer.
	Mesh *mesh = new Mesh();

	if (mSubMesh[0].mSharedVertexBuffer)
	{
		*mesh	=	*this;
		
		return	mesh;
	}
	
	mesh->mVertexBuffer	=	new VertexBuffer();
	mesh->mSubMeshCount =	mSubMeshCount;
	
	// Reserva memoria para los submeshes
	mesh->mSubMesh	=	new SubMesh[mSubMeshCount];

	//	Fill up bounding box settings.
	mesh->mMeshBounds.maxX					=	mMeshBounds.maxX;
	mesh->mMeshBounds.maxY					=	mMeshBounds.maxY;
	mesh->mMeshBounds.maxZ					=	mMeshBounds.maxZ;
	mesh->mMeshBounds.minX					=	mMeshBounds.minX;
	mesh->mMeshBounds.minY					=	mMeshBounds.minY;
	mesh->mMeshBounds.minZ					=	mMeshBounds.minZ;
	mesh->mMeshBounds.radius				=	mMeshBounds.radius;
	mesh->mMeshBounds.scaleFactor		=	mMeshBounds.scaleFactor;
	
// construccin de los submeshes
	long int acumVerts	=	0;
	
	for (size_t i = 0; i < mesh->mSubMeshCount; i++)
	{
		mesh->mSubMesh[i].mSharedVertexBuffer	=	true;
		mesh->mSubMesh[i].mVertexBuffer				=	mesh->mVertexBuffer;
		mesh->mSubMesh[i].mStripCount					=	0;
		mesh->mSubMesh[i].mStrip							=	NULL;
		mesh->mSubMesh[i].mType								=	mSubMesh[i].mType;
		
		strcpy(mesh->mSubMesh[i].mMaterialName,mSubMesh[i].mMaterialName);
		
		// copiar los ndices
		mesh->mSubMesh[i].mIndexCount	=	mSubMesh[i].mIndexCount;
		mesh->mSubMesh[i].mIndex			=	new Index[mSubMesh[i].mIndexCount];

		for (size_t j = 0;	j < mSubMesh[i].mIndexCount;	j++)
		{ 
			mesh->mSubMesh[i].mIndex[j]	=	mSubMesh[i].mIndex[j]+acumVerts;
		}

		acumVerts	+=	long(mSubMesh[i].mVertexBuffer->mVertexCount);

		// Copiar las tiras
		int offset	=	0; // desplazamiento entre posiciones de memoria.
		
		mesh->mSubMesh[i].mStripCount	=	mSubMesh[i].mStripCount;
		
		if (mSubMesh[i].mStripCount > 0)
		{
			mesh->mSubMesh[i].mStrip	=	new Index*[mSubMesh[i].mStripCount];

			offset	=	int(&(mesh->mSubMesh[i].mIndex[0]) - &(mSubMesh[i].mIndex[0]));
			
			for (size_t j = 0;	j < mSubMesh[i].mStripCount;	j++)
			{
				mesh->mSubMesh[i].mStrip[j] = mSubMesh[i].mStrip[j] + offset;
			}
		}
	}
	
	mesh->mVertexBuffer->mVertexCount	=	acumVerts;
	mesh->mVertexBuffer->mVertexInfo	=	mSubMesh[0].mVertexBuffer->mVertexInfo;
	
	mesh->mVertexBuffer->mPosition		=	new Vector3[mesh->mVertexBuffer->mVertexCount];
	mesh->mVertexBuffer->mNormal			=	new Vector3[mesh->mVertexBuffer->mVertexCount];
	mesh->mVertexBuffer->mTexCoords		=	new Vector2[mesh->mVertexBuffer->mVertexCount];

	// copiar los vrtices
	acumVerts	=	0;
	
	size_t	newIndex;
	
	for (size_t i = 0; i < mSubMeshCount;	i++)
	{
		for(size_t j = 0; j < mSubMesh[i].mVertexBuffer->mVertexCount; j++)
		{
			newIndex	=	acumVerts + j;
			
			mesh->mVertexBuffer->
						mPosition[newIndex].x	=	mSubMesh[i].mVertexBuffer->
																								mPosition[j].x;
			
			mesh->mVertexBuffer->
						mPosition[newIndex].y	=	mSubMesh[i].mVertexBuffer->
																								mPosition[j].y;
			
			mesh->mVertexBuffer->
						mPosition[newIndex].z	=	mSubMesh[i].mVertexBuffer->
																								mPosition[j].z;

			mesh->mVertexBuffer->
						mNormal[newIndex].x	=	mSubMesh[i].mVertexBuffer->
																							mNormal[j].x;
			
			mesh->mVertexBuffer->
						mNormal[newIndex].y	=	mSubMesh[i].mVertexBuffer->
																							mNormal[j].y;
			
			mesh->mVertexBuffer->
						mNormal[newIndex].z	=	mSubMesh[i].mVertexBuffer->
																							mNormal[j].z;

			mesh->mVertexBuffer->
						mTexCoords[newIndex].x	=	mSubMesh[i].mVertexBuffer->
																									mTexCoords[j].x;
			
			mesh->mVertexBuffer->
						mTexCoords[newIndex].y	=	mSubMesh[i].mVertexBuffer->
																									mTexCoords[j].y;
		}
		
		acumVerts	+=	long(mSubMesh[i].mVertexBuffer->mVertexCount);

		//	Copy submesh bones.
		if (!mSubMesh[i].mBones.empty())
		{
			for (unsigned int j = 0; j < mSubMesh[i].mBones.size(); j++)
			{
				mesh->mSubMesh[i].mBones.push_back(mSubMesh[i].mBones[j]);
			}
		}

	}

	//	Copy skeleton name.
	if (hasSkeleton)
	{
		mesh->hasSkeleton	=	true;
		
		strcpy(mesh->mSkeletonName,mSkeletonName);
	}

	//	Copy mesh bones.
	if (!mBones.empty())
	{
		for (unsigned int j = 0; j < mBones.size(); j++)
		{
			mesh->mBones.push_back(mBones[j]);
		}
	}

	return mesh;
}

